-- Lua optimizations
local date = date
local gsub = gsub
local select = select
local strfind = strfind
local strjoin = strjoin
local strsplit = strsplit
local strupper = strupper
local tonumber = tonumber
local wipe = wipe
local GetItemInfo = GetItemInfo
local GetRealmName = GetRealmName
local SecondsToTime = SecondsToTime
local UnitClass = UnitClass
local UnitName = UnitName

-- Use Blizzard's strings to save localization efforts
local NO_RAID_INSTANCES_SAVED = NO_RAID_INSTANCES_SAVED -- "You are not saved to any instances"
local FONTEND = FONT_COLOR_CODE_CLOSE
local NORMALFONT = NORMAL_FONT_COLOR_CODE
local YELLOWFONT = LIGHTYELLOW_FONT_COLOR_CODE
local GRAYFONT = GRAY_FONT_COLOR_CODE

local EggTimer = LibStub("AceAddon-3.0"):NewAddon("EggTimer", "AceEvent-3.0", "AceTimer-3.0", "LibSink-2.0")
_G.EggTimer = EggTimer
local Config = LibStub("AceConfig-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("EggTimer")

local LSM = LibStub("LibSharedMedia-3.0", true)
if not LSM then
	LoadAddOn("LibSharedMedia-3.0")
	LSM = LibStub("LibSharedMedia-3.0", true)
end
local DEFAULT_LSM_FONT = "Arial Narrow"
if LSM then
	if not LSM:IsValid("font", DEFAULT_LSM_FONT) then
		-- non-Western languages
		DEFAULT_LSM_FONT = LSM:GetDefault("font")
	end
end

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)
local launcher

local QTip = LibStub("LibQTip-1.0")
local tooltip

local bars
local feedTimer
local expiryTimers = { }

local thisToon = strjoin(" - ", UnitName("player"), GetRealmName())

local db
local defaultDB = {
	DBVersion = 2,
	Toons = { },-- table key: "Toon - Realm"; value:
					-- Class: string
					-- Hide: boolean
					-- ConsortiumShow: boolean (false) -- show Consortium alerts for this toon
					-- ConsortiumLastMonth: boolean (false) -- last month that the toon got his gems
					-- ConsortiumRep: boolean (false) -- this toon has enough Consortium reputation
	Items = { },-- table key: item code; value:
					-- table key: "Toon - Realm"; value:
						-- Expires: integer
	Bars = {
		Single = false, -- show only the most imminent bar
		Anchor = "CENTER", -- bar frame anchor
		GrowUp = false,
		Locked = true,
		Icon = true,
		Texture = "Blizzard",
		ColorText = {
			R = 1,
			G = 1,
			B = 1,
			A = 1,
		},
		ColorBar = {
			R = 0,
			G = 0,
			B = 1,
			A = 1,
		},
		ColorBarBG = {
			R = 0,
			G = 0,
			B = 1,
			A = 0.3,
		},
		PosX = 0,
		PosY = 0,
		Height = 12,
		Width = 160,
	},
	Calendar = {
		Enable = false,
	},
	MinimapIcon = { },
	Announce = {
		Sound = false,
		Message = false,
		SoundFile = "None",
		SinkOptions = { },
	},
}

-- general helper functions below

local function ColorCodeOpen(color)
	return format("|c%02x%02x%02x%02x", math.floor(color[4] * 255), math.floor(color[1] * 255), math.floor(color[2] * 255), math.floor(color[3] * 255))
end

local function ClassColorise(class, targetstring)
	local color = {
		RAID_CLASS_COLORS[class].r,
		RAID_CLASS_COLORS[class].g,
		RAID_CLASS_COLORS[class].b,
		1,
	}
	return strjoin("", ColorCodeOpen(color), targetstring, FONTEND)
end

local function TableLen(table)
	local i = 0
	for k, v in pairs(table) do
		i = i + 1
	end
	return i
end

-- local addon functions below

local function CullStaleItems()
	for item, a in pairs(EggTimer.db.Items) do
		for toon, b in pairs(a) do
			if b.Expires < time() then
				EggTimer.db.Items[item][toon] = nil
			end
		end
	end
	for item, a in pairs(EggTimer.db.Items) do
		if TableLen(a) == 0 then
			EggTimer.db.Items[item] = nil
		end
	end
end

local function HideTooltips()
	if not tooltip then return end
	QTip:Release(tooltip)
	tooltip = nil
end

local function ShowTooltip(anchorframe)
	if tooltip then QTip:Release(tooltip) end
	tooltip = QTip:Acquire("EggTimerTooltip", 2, "LEFT")
	local hFont = tooltip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	tooltip:SetHeaderFont(hFont)
	CullStaleItems()
	local toons = EggTimer.db.Toons
	local columns = { }
	if not toons[thisToon].Hide and toons[thisToon].ConsortiumShow and toons[thisToon].ConsortiumRep and toons[thisToon].ConsortiumLastMonth ~= CalendarGetMonth() then
		columns[thisToon] = columns[thisToon] or tooltip:AddColumn("CENTER")
	end
	local headLine, _ = tooltip:AddHeader(NORMALFONT .. "EggTimer" .. FONTEND)
	-- do Consortium stuff
	local consortiumLine
	for toon, a in pairs(EggTimer.db.Toons) do
		if a.ConsortiumShow and a.ConsortiumRep and a.ConsortiumLastMonth < CalendarGetMonth() then
			columns[toon] = columns[toon] or tooltip:AddColumn("CENTER")
			if not consortiumLine then
				consortiumLine = tooltip:AddLine()
				tooltip:SetCell(consortiumLine, 1, L["this month's gems"], "LEFT")
			end
			tooltip:SetCell(consortiumLine, columns[toon], L["waiting"], "CENTER")
		end
	end
	-- do other items
	for item, a in pairs(EggTimer.db.Items) do
		local itemLine, _ = tooltip:AddLine()
		tooltip:SetCell(itemLine, 1, GetItemInfo(item), "LEFT")
		for toon, b in pairs(a) do
			columns[toon] = columns[Toon] or tooltip:AddColumn("CENTER")
			tooltip:SetCell(itemLine, columns[toon], SecondsToTime(b.Expires - time()), "RIGHT")
		end
	end
	for toon, col in pairs(columns) do
		tooltip:SetCell(headLine, col, ClassColorise(EggTimer.db.Toons[toon].Class, strsplit(' ', toon)), tooltip:GetHeaderFont(), "CENTER")
	end 
	tooltip:AddSeparator(8,0,0,0,0)
	local hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, "|cffffff00".."Click|r to configure", "LEFT", tooltip:GetColumnCount()) 
	tooltip:SmartAnchorTo(anchorframe)
	tooltip:Show()
end

-- global addon code below

function EggTimer:OnInitialize()
	EggTimerDB = EggTimerDB or defaultDB
	if not EggTimerDB.DBVersion then
		EggTimerDB = defaultDB
	end
	self.expiryTimers = { }
	self.db = db or EggTimerDB
	self.db.Toons[thisToon] = self.db.Toons[thisToon] or { }
	self.db.Toons[thisToon].Class = select(2, UnitClass("player"))
	self.db.Toons[thisToon].Hide = self.db.Toons[thisToon].Hide or false
	self.db.Toons[thisToon].ConsortiumShow = self.db.Toons[thisToon].ConsortiumShow or true
	self.db.Toons[thisToon].ConsortiumRep = self.db.Toons[thisToon].ConsortiumRep or false
	self:SetupOptions()
	self:SetSinkStorage(EggTimer.db.Announce.SinkOptions)
--	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "UpdateUsedMedia")
	if LDB then
		self.launcher = LDB:NewDataObject("EggTimer", {
			label = "EggTimer",
			type = "data source",
			icon = "Interface\\Icons\\INV_Egg_02",
			text = L["No perishables"],
			OnClick = function(clickedframe, button)
				EggTimer:ShowConfig()
			end,
			OnEnter = function(self)
				ShowTooltip(self)
			end,
			OnLeave = function(self)
				HideTooltips()
			end,
		})
		self.feedTimer = self:ScheduleRepeatingTimer("UpdateFeed", 5, nil)
	end
	if LDBIcon then
		LDBIcon:Register("EggTimer", self.launcher, self.db.MinimapIcon)
	end
	self:UpdateFeed()
	self:PrepareAnnouncements()
end

function EggTimer:OnEnable()
	self:RegisterEvent("TRADE_SHOW")
	self:RegisterEvent("TRADE_CLOSED")
	self:RegisterEvent("CHAT_MSG_LOOT", "OnReceiveItem")
end

function EggTimer:OnDisable()
	self:UnregisterEvent("TRADE_SHOW")
	self:UnregisterEvent("TRADE_CLOSED")
	self:UnregisterEvent("CHAT_MSG_LOOT")
	self:CancelTimer(feedTimer, true)
	for i, v in ipairs(self.expiryTimers) do
		self:CancelTimer(v, true)
	end
end

function EggTimer:TRADE_SHOW()
	self:UnregisterEvent("CHAT_MSG_LOOT")
end

function EggTimer:TRADE_CLOSED()
	self:RegisterEvent("CHAT_MSG_LOOT", "OnReceiveItem")
end

function EggTimer:UpdateFeed()
	local nextexpiry
	for item, a in pairs(EggTimer.db.Items) do
		for toon, b in pairs(a) do
			if not nextexpiry or EggTimer.db.Items[item][toon].Expires < nextexpiry then
				nextexpiry = EggTimer.db.Items[item][toon].Expires
			end
		end
	end
	if nextexpiry then
		self.launcher.text = SecondsToTime(nextexpiry - time())
	else
		self.launcher.text = "No perishables"
	end
end

function EggTimer:OnReceiveItem(eventName, arg1)
	local item
	_, _, item = strfind(arg1, gsub(LOOT_ITEM_PUSHED_SELF_MULTIPLE, "(%%s)", "(.+)"))
	if not item then
		_, _, item = strfind(arg1, gsub(LOOT_ITEM_PUSHED_SELF, "(%%s)", "(.+)"))
	end
	if item then
		local _, itemlink = GetItemInfo(item)
		local _, _, itemcode = strfind(itemlink, "(%d+):")
		itemcode = tonumber(itemcode)
		-- Consortium checks
		if itemcode == 25424 or itemcode == 25419 or itemcode == 25422 or itemcode == 25423 then
			EggTimer.db.Toons[thisToon].ConsortiumRep = true
			EggTimer.db.Toons[thisToon].ConsortiumLastMonth = CalendarGetMonth()
		-- other item checks
		elseif EggTimer.ItemDB[itemcode] then
			if EggTimer.ItemDB[itemcode].StartItem then
				local initialitem = EggTimer.ItemDB[itemcode].StartItem
				if EggTimer.db.Items[initialitem] then
					if EggTimer.db.Items[initialitem][thisToon] then
						EggTimer.db.Items[initialitem][thisToon] = nil
					end
					if TableLen(EggTimer.db.Items[initialitem]) == 0 then
						EggTimer.db.Items[initialitem] = nil
					end
				end
				local itemname = GetItemInfo(initalitem)
				if not itemname then -- server will need to be queried with a fake tooltip
					GameTooltip:Hide()
					GameTooltip:SetHyperlink("item:"..initialitem)
					EggTimer:ScheduleTimer("OnReceiveItem", 0.5, eventName, arg1)
					return
				end
				local message
				if EggTimer.db.Announce.Message then
					message = format(L["expiry message format"], itemname)
				else
					message = false
				end
				EggTimer:Announce(EggTimer.db.Announce.Sound, message)
			else
				EggTimer.db.Items[itemcode] = EggTimer.db.Items[itemcode] or { }
				if EggTimer.db.Items[itemcode][thisToon] then
					wipe(EggTimer.db.Items[itemcode][thisToon])
				else
					EggTimer.db.Items[itemcode][thisToon] = { }				
				end
				EggTimer.db.Items[itemcode][thisToon].Expires = time() + EggTimer.ItemDB[itemcode].Duration
				if not EggTimer.ItemDB[itemcode].EndItem then
					local message
					if EggTimer.db.Announce.Message then
						message = format(L["expiry message format"], GetItemInfo(itemcode))
					else
						message = false
					end
					tinsert(self.expiryTimers, self:ScheduleTimer("Announce", EggTimer.ItemDB[itemcode].Duration, EggTimer.db.Announce.Sound, message))
				end
				if self.db.Calendar.Enable and EggTimer.ItemDB[itemcode].Calendar then
					self:ShowCalendarPrompt(itemcode)
				end
			end
		end
	end 
end

function EggTimer:PrepareAnnouncements()
	for item, a in pairs(EggTimer.db.Items) do
		for toon, b in pairs(a) do
			if not (toon == thisToon and EggTimer.ItemDB[item].EndItem) then
				local itemname = GetItemInfo(item)
				if not itemname then -- server will need to be queried with a fake tooltip
					GameTooltip:Hide()
					GameTooltip:SetHyperlink("item:"..item)
					EggTimer:ScheduleTimer("PrepareAnnouncements", 0.5)
					return
				end
				local message
				if EggTimer.db.Announce.Message then
					local itemname = GetItemInfo(item)
					message = format(L["expiry message format"], itemname)
					message = message .. " (" .. ClassColorise(EggTimer.db.Toons[toon].Class, strsplit(" ", toon)) .. ")"
				else
					message = false
				end
				tinsert(EggTimer.expiryTimers, EggTimer:ScheduleTimer("Announce", EggTimer.ItemDB[item].Duration, EggTimer.db.Announce.Sound, message))
			end
		end
	end
end

function EggTimer:ShowCalendarPrompt(itemcode)
	EggTimer.calendaritems = EggTimer.calendaritems or { }
	EggTimer.calendaritems[#EggTimer.calendaritems+1] = itemcode
	EggTimerCalendarPrompt:Show()
end

function EggTimer:Announce(sound, message)
	if sound then PlaySoundFile(LSM:Fetch("sound", EggTimer.db.Announce.SoundFile)) end
	if message then self:Pour(message, 1, 1, 1, nil, 24, "OUTLINE", true) end
end

-- options stuff below

EggTimer.Options = {
	type = "group",
	name = "EggTimer",
	handler = EggTimer,
	get = function(info) return EggTimer.db[ info[#info] ] end,
	set = function(info, value) EggTimer.db[ info[#info] ] = value end,
	args = {
		General = {
			order = 1,
			type = "group",
			name = "General Settings",
			desc = "General Settings",
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Monitors the time left on your Mysterious Egg and other perishable items",
				},
				MinimapIcon = {
					type = "toggle",
					name = "Show minimap button",
					desc = "Show the EggTimer minimap button",
					order = 2,
					get = function(info) return not EggTimer.db.MinimapIcon.hide end,
					set = function(info, value)
						EggTimer.db.MinimapIcon.hide = not value
						if value then LDBIcon:Show("EggTimer") else LDBIcon:Hide("EggTimer") end
					end,
					hidden = function() return not LDBIcon end,
				},
				Calendar = {
					order = 3,
					type = "toggle",
					name = "Enable Calendar events",
					desc = "Create new Blizzard Calendar events automatically",
					get = function(info) return EggTimer.db.Calendar.Enable end,
					set = function(info, v)
						EggTimer.db.Calendar.Enable = v
						return EggTimer.db.Calendar.Enable
					end,
				},
				ConsortiumNote = {
					order = 4,
					type = "description",
					name = L["EggTimer will track your Consortium gem pouches when you next receive them"],
					hidden = function()
						local toonfound = false
						for k, v in pairs(EggTimer.db.Toons) do
							if v.ConsortiumRep then
								toonfound = true
							end
						end
						return toonfound
					end,
				},
				Consortium = {
					order = 4,
					type = "multiselect",
					name = L["Consortium gem reminders"],
					desc = L["Show monthly reminder for gems from The Consortium for these characters only"],
					get = function(info, toon)
						return EggTimer.db.Toons[toon].ConsortiumShow
					end,
					set = function(info, toon, value)
						EggTimer.db.Toons[toon].ConsortiumShow = value
					end,
					values = function()
						local list = { }
						local k, v
						for k, v in pairs(EggTimer.db.Toons) do
							if v.ConsortiumRep then
								list[k] = k
							end
						end
						return list
					end,
					hidden = function()
						local toonfound = false
						for k, v in pairs(EggTimer.db.Toons) do
							if v.ConsortiumRep then
								toonfound = true
							end
						end
						return not toonfound
					end,
				},
			},
		},
		Bar = {
			order = 3,
			type = "group",
			name = "Bar Settings",
			desc = "Configure timer bar display",
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Configure timer bar display",
				},
			},
		},
		Tooltip = {
			order = 4,
			type = "group",
			name = "Tooltip Settings",
			desc = "Select the contents of the tooltip",
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Select the contents of the tooltip",
				},
			},
		},
		Announce = {
			order = 5,
			type = "group",
			name = "Announcements",
			desc = "Choose your notification method",
			get = function(info) return EggTimer.db.Announce[ info[#info] ] end,
			set = function(info, value)
				EggTimer.db.Announce[ info[#info] ] = value
			end,
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Choose your notification method",
				},
				Sound = {
					type = "toggle",
					order = 2,
					name = "Enable Sound",
					desc = "Causes EggTimer to play a chosen sound effect",
				},
				Message = {
					type = "toggle",
					order = 3,
					name = "Enable Announcement",
					desc = "Print a message to screen when items expire",
				},
				Output = EggTimer:GetSinkAce3OptionsDataTable(),
				SoundFile = {
					type = "select", dialogControl = 'LSM30_Sound',
					order = 5,
					name = "Sound to play",
					desc = "Sound to play",
					values = AceGUIWidgetLSMlists.sound,
					disabled = function() return not EggTimer.db.Announce.Sound end,
				},
				test = {
					type = "execute",
					order = -1,
					name = "Test announcement",
					desc = "Test announcement",
					func = function()
						EggTimer:Announce(EggTimer.db.Announce.Sound, EggTimer.db.Announce.Message and "Test announcement")
					end,
				},
			},
		},
	},
}

local optionsFrames = { }
function EggTimer:SetupOptions()
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("EggTimer", EggTimer.Options)
	optionsFrames["General"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("EggTimer", nil, nil, "General")
--	optionsFrames["Bar"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("EggTimer", L["Bar Settings"], "EggTimer", "Bar")
--	optionsFrames["Tooltip"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("EggTimer", L["Tooltip Settings"], "EggTimer", "Tooltip")
	optionsFrames["Announce"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("EggTimer", L["Announcements"], "EggTimer", "Announce")
end

function EggTimer:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(optionsFrames["Announce"])
	InterfaceOptionsFrame_OpenToCategory(optionsFrames["General"])
end

